import difflib
import sys
import types
from dataclasses import dataclass, field, fields
from enum import Enum
from inspect import isclass
from itertools import chain
from typing import Any, Optional, Type, TypedDict, Union, get_args, get_origin

import pyparsing as pp

from ..enums import AccountEdition, DataType, ParseableEnum, ResourceType
from ..exceptions import ResourceHasContainerException, WrongContainerException, WrongEditionException
from ..identifiers import FQN, URN, parse_identifier, resource_label_for_type
from ..lifecycle import create_resource, drop_resource
from ..parse import _parse_create_header, _parse_props, resolve_resource_class
from ..props import Props as ResourceProps
from ..resource_name import ResourceName
from ..resource_tags import ResourceTags
from ..role_ref import RoleRef
from ..scope import (
    AccountScope,
    DatabaseScope,
    OrganizationScope,
    ResourceScope,
    SchemaScope,
    resource_can_be_contained_in,
)
from ..var import VarString, string_contains_var


def _suggest_correct_kwargs(expected_kwargs, passed_kwargs):
    suggestions = {}
    for passed_kwarg in passed_kwargs:
        # Find the closest match from the expected kwargs for each passed kwarg
        closest_match = difflib.get_close_matches(passed_kwarg, expected_kwargs, n=1, cutoff=0.6)
        if closest_match:
            closest_match = closest_match[0]
            # If the passed kwarg is not exactly an expected kwarg, add it to suggestions
            if passed_kwarg != closest_match:
                suggestions[passed_kwarg] = closest_match
        else:
            # If no close match is found, suggest it might be an unexpected kwarg
            suggestions[passed_kwarg] = "Unexpected kwarg, no close match found."

    return suggestions


class Arg(TypedDict):
    name: str
    data_type: DataType


class Returns(TypedDict):
    data_type: DataType
    metadata: str


@dataclass
class ResourceLifecycleConfig:
    ignore_changes: list[str] = field(default_factory=list)
    prevent_destroy: bool = False


def _coerce_resource_field(field_value, field_type):

    # No type checking or coercion for Any
    if field_type == Any:
        return field_value

    # Recursively traverse lists and dicts
    elif get_origin(field_type) is list:
        if not isinstance(field_value, list):
            raise TypeError
        list_element_type = get_args(field_type) or (str,)
        return [_coerce_resource_field(v, field_type=list_element_type[0]) for v in field_value]

    elif get_origin(field_type) is dict:
        if not isinstance(field_value, dict):
            raise TypeError
        dict_types = get_args(field_type)
        if len(dict_types) < 2:
            raise RuntimeError(f"Unexpected field type {field_type}")
        return {k: _coerce_resource_field(v, field_type=dict_types[1]) for k, v in field_value.items()}

    elif field_type is RoleRef:
        return convert_role_ref(field_value)

    # Check for field_value's type in a Union
    elif get_origin(field_type) == Union:
        union_types = get_args(field_type)
        for union_type in union_types:
            expected_type = get_origin(union_type) or union_type
            if isinstance(field_value, expected_type):
                return _coerce_resource_field(field_value, field_type=expected_type)
        raise RuntimeError(f"Unexpected field type {field_type}")

    elif not isclass(field_type):
        raise RuntimeError(f"Unexpected field type {field_type}")

    # Coerce enums
    elif issubclass(field_type, ParseableEnum):
        try:
            new_value = field_type(field_value)
        except ValueError:
            raise TypeError
        return new_value

    # Coerce args
    elif field_type is Arg:
        arg_dict = {
            "name": field_value["name"].upper(),
            "data_type": DataType(field_value["data_type"]),
        }
        if "default" in field_value:
            arg_dict["default"] = field_value["default"]
        return arg_dict

    # Coerce returns
    elif field_type is Returns:
        returns_dict = {
            "data_type": DataType(field_value["data_type"]),
            "metadata": field_value["metadata"],
        }
        if "returns_null" in field_value:
            returns_dict["returns_null"] = field_value["returns_null"]
        return returns_dict

    # Coerce resources
    elif issubclass(field_type, Resource):
        return convert_to_resource(field_type, field_value)
    elif field_type is ResourceName:
        return field_value if isinstance(field_value, VarString) else ResourceName(field_value)
    elif field_type is ResourceTags:
        return ResourceTags(field_value)
    elif field_type is str:
        if isinstance(field_value, str) and string_contains_var(field_value):
            return VarString(field_value)
        elif isinstance(field_value, VarString):
            return field_value
        elif not isinstance(field_value, str):
            raise TypeError
        else:
            return field_value
    elif field_type is float:
        if isinstance(field_value, float):
            return field_value
        elif isinstance(field_value, int):
            return float(field_value)
        else:
            raise TypeError
    else:
        # Typecheck all other field types (str, int, etc.)
        if not isinstance(field_value, field_type):
            raise TypeError
        return field_value


@dataclass
class ResourceSpecMetadata:
    fetchable: bool = True
    triggers_replacement: bool = False
    triggers_create: bool = False
    ignore_changes: bool = False
    known_after_apply: bool = False
    edition: set[AccountEdition] = field(
        default_factory=lambda: {AccountEdition.STANDARD, AccountEdition.ENTERPRISE, AccountEdition.BUSINESS_CRITICAL}
    )


@dataclass
class ResourceSpec:

    def to_dict(self, account_edition: AccountEdition):
        dict_: dict[str, Any] = {}

        def _serialize_field(field, value):
            if field.name == "owner":
                return str(value.fqn)
            elif isinstance(value, ResourcePointer):
                return str(value.fqn)
            elif isinstance(value, Resource):
                if getattr(value, "serialize_inline", False):
                    return value.to_dict(account_edition)
                elif isinstance(value, NamedResource):
                    return str(value.fqn)
                else:
                    raise Exception(f"Cannot serialize {value}")
            elif isinstance(value, ParseableEnum):
                return str(value)
            elif isinstance(value, list):
                return [_serialize_field(field, v) for v in value]
            elif isinstance(value, dict):
                return {k: _serialize_field(field, v) for k, v in value.items()}
            elif isinstance(value, ResourceName):
                return str(value)
            elif isinstance(value, ResourceTags):
                return value.tags
            else:
                return value

        for f in fields(self):
            value = getattr(self, f.name)
            field_metadata = ResourceSpecMetadata(**f.metadata)
            if account_edition not in field_metadata.edition:
                if value != f.default and value is not None:
                    raise WrongEditionException(
                        f"Field {self.__class__.__name__}.{f.name} is not supported in edition {account_edition}. Supported editions: {field_metadata.edition}"
                    )
                else:
                    continue
            dict_[f.name] = _serialize_field(f, value)

        return dict_

    def __post_init__(self):
        for f in fields(self):
            field_value = getattr(self, f.name)
            if field_value is None:
                continue
            else:
                try:
                    new_value = _coerce_resource_field(field_value, f.type)
                    setattr(self, f.name, new_value)
                except TypeError as err:
                    human_readable_classname = self.__class__.__name__[1:]
                    if isclass(f.type) and issubclass(f.type, Enum):
                        raise TypeError(
                            f"Expected {human_readable_classname}.{f.name} to be one of ({', '.join(f.type.__members__.keys())}), got {repr(field_value)} instead"
                        ) from err
                    else:
                        raise TypeError(
                            f"Expected {human_readable_classname}.{f.name} to be {f.type}, got {repr(field_value)} instead"
                        ) from err

    @classmethod
    def get_metadata(cls, field_name: str) -> ResourceSpecMetadata:
        for f in fields(cls):
            if f.name == field_name:
                return ResourceSpecMetadata(**f.metadata)
        raise ValueError(f"Field {field_name} not found in {cls.__name__}")


RESOURCE_SCOPES = {
    ResourceType.ACCOUNT: OrganizationScope(),
}


class _Resource(type):
    __types__: dict[ResourceType, list[Type["Resource"]]] = {}
    __classes__: dict[str, Type["Resource"]] = {}
    __resolvers__ = {}

    def __new__(cls, name, bases, attrs):
        cls_ = super().__new__(cls, name, bases, attrs)
        if cls_.__name__ in ["Resource", "_Resource", "ResourcePointer"]:
            return cls_
        cls.__classes__[name] = cls_
        if cls_.resource_type not in cls.__types__:
            cls.__types__[cls_.resource_type] = []
        cls.__types__[cls_.resource_type].append(cls_)
        if cls_.resource_type not in RESOURCE_SCOPES:
            RESOURCE_SCOPES[cls_.resource_type] = cls_.scope
        return cls_


class Resource(metaclass=_Resource):
    edition = {AccountEdition.STANDARD, AccountEdition.ENTERPRISE, AccountEdition.BUSINESS_CRITICAL}
    props: ResourceProps
    resource_type: ResourceType
    scope: ResourceScope
    spec: Type[ResourceSpec]
    serialize_inline: bool = False

    def __init__(
        self,
        implicit: bool = False,
        lifecycle: dict = None,
        **kwargs,
    ):
        super().__init__()
        self._data: ResourceSpec = None
        self._container: "ResourceContainer" = None
        self._finalized = False
        self.lifecycle = ResourceLifecycleConfig(**lifecycle) if lifecycle else ResourceLifecycleConfig()
        self.implicit = implicit
        self.refs: list[Resource] = []

        # Consume resource_type from kwargs if it exists
        resource_type = kwargs.pop("resource_type", None)
        resource_type = ResourceType(resource_type) if resource_type else None
        if resource_type and resource_type != self.resource_type:
            raise ValueError(f"Unexpected resource_type {resource_type} for {self.resource_type}")

        # Consume scope from kwargs if it exists
        database = kwargs.pop("database", None)
        schema = kwargs.pop("schema", None)
        self._register_scope(database=database, schema=schema)

        # If there are more kwargs, throw an error
        # Based on https://stackoverflow.com/questions/1603940/how-can-i-modify-a-python-traceback-object-when-raising-an-exception
        if kwargs:
            try:
                if self.spec:
                    field_names = [f.name for f in fields(self.spec)]
                    field_names = ", ".join(field_names)
                    raise ValueError(
                        f"Unexpected keyword arguments for {self.__class__.__name__} {kwargs}. Valid field names: {field_names}"
                    )
                else:
                    raise ValueError(f"Unexpected keyword arguments for {self.__class__.__name__} {kwargs}")
            except ValueError as err:
                traceback = sys.exc_info()[2]
                back_frame = traceback.tb_frame.f_back
                msg = str(err)

            back_tb = types.TracebackType(
                tb_next=None, tb_frame=back_frame, tb_lasti=back_frame.f_lasti, tb_lineno=back_frame.f_lineno
            )
            raise ValueError(msg).with_traceback(back_tb)

    @classmethod
    def from_sql(cls, sql):
        resource_cls = cls
        if resource_cls == Resource:
            # FIXME: we need to change the way we handle polymorphic resources
            # make a new function called _parse_resource_type_from_create
            # resource_cls = Resource.classes[_resolve_resource_class(sql)]
            # raise NotImplementedError
            resource_type = resolve_resource_class(sql)
            scope = RESOURCE_SCOPES[resource_type]
        else:
            resource_type = resource_cls.resource_type
            scope = resource_cls.scope

        identifier, remainder_sql = _parse_create_header(sql, resource_type, scope)

        try:
            props = _parse_props(resource_cls.props, remainder_sql) if remainder_sql else {}
            return resource_cls(**identifier, **props)
        except pp.ParseException as err:
            raise pp.ParseException(f"Error parsing {resource_cls.__name__} props {identifier}") from err

    @classmethod
    def from_dict(cls, data: dict):
        resource_cls = cls.resolve_resource_cls(ResourceType(data["resource_type"]), data)
        return resource_cls(**data)

    @classmethod
    def props_for_resource_type(cls, resource_type: ResourceType, data: dict = None):
        return cls.resolve_resource_cls(resource_type, data).props

    @classmethod
    def resolve_resource_cls(cls, resource_type: ResourceType, data: dict = None) -> Type["Resource"]:

        if not isinstance(resource_type, ResourceType):
            raise ValueError(f"Expected ResourceType, got {resource_type}({type(resource_type)})")
        if resource_type not in cls.__types__:
            raise ValueError(f"Resource class for type not found {resource_type}")
        resource_types = cls.__types__[resource_type]

        if len(resource_types) > 1:
            if data is None:
                raise ValueError(f"Cannot resolve polymorphic resource class [{resource_type}] without data")
            else:
                resolver = cls.__resolvers__[resource_type]
                return resolver(data)
        return resource_types[0]

    @classmethod
    def defaults(cls):
        return {f.name: f.default for f in fields(cls.spec)}

    def __repr__(self):  # pragma: no cover
        if not hasattr(self, "_data"):
            return f"{self.__class__.__name__}(<uninitialized>)"
        name = getattr(self._data, "name", "<noname>")
        implicit = "~" if self.implicit else ""
        return f"{self.__class__.__name__}({implicit}{name})"

    def __eq__(self, other):
        if not isinstance(other, Resource):
            return False
        return self._data == other._data

    def __hash__(self):
        return hash(URN.from_resource(self, ""))

    def to_dict(self, account_edition: Optional[AccountEdition] = None):
        return self._data.to_dict(account_edition or AccountEdition.ENTERPRISE)

    def create_sql(
        self,
        account_edition: Optional[AccountEdition] = None,
        **kwargs,
    ):
        return create_resource(
            self.urn,
            self.to_dict(account_edition),
            self.props,
            **kwargs,
        )

    def drop_sql(
        self,
        if_exists: bool = False,
        account_edition: Optional[AccountEdition] = None,
    ):
        return drop_resource(self.urn, self.to_dict(account_edition), if_exists=if_exists)

    def _requires(self, resource: "Resource"):
        if self._finalized:
            raise RuntimeError("Cannot modify a finalized resource")
        if isinstance(resource, (Resource, ResourcePointer)):
            self.refs.append(resource)

    def requires(self, *resources: "Resource"):
        if isinstance(resources[0], list):
            resources = resources[0]
        for resource in resources:
            self._requires(resource)

    def _register_scope(self, database=None, schema=None):
        if isinstance(database, str):
            database = ResourcePointer(name=database, resource_type=ResourceType.DATABASE)

        if isinstance(schema, str):
            schema = ResourcePointer(name=schema, resource_type=ResourceType.SCHEMA)
            if database is not None:
                database.add(schema)

        if isinstance(self.scope, DatabaseScope):
            if schema is not None:
                raise RuntimeError(f"Unexpected kwarg schema {schema} for resource {self}")
            if database is not None:
                database.add(self)

        elif isinstance(self.scope, SchemaScope):
            if schema is not None:
                schema.add(self)
                if database is not None:
                    if schema.container is None:
                        database.add(schema)
                    elif schema.container != database:
                        raise ResourceHasContainerException(f"Schema {schema} does not belong to database {database}")

            elif database is not None:
                database.find(name="PUBLIC", resource_type=ResourceType.SCHEMA).add(self)

    def _resolve_vars(self, vars: dict):

        def _render_vars(field_value):
            if isinstance(field_value, VarString):
                return field_value.to_string(vars)
            elif isinstance(field_value, list):
                return [_render_vars(v) for v in field_value]
            elif isinstance(field_value, dict):
                return {k: _render_vars(v) for k, v in field_value.items()}
            elif isinstance(field_value, Resource) and not isinstance(field_value, ResourcePointer):
                field_value._resolve_vars(vars)
                return field_value
            else:
                return field_value

        if self._data:
            for f in fields(self._data):
                field_value = getattr(self._data, f.name)
                new_value = _render_vars(field_value)
                setattr(self._data, f.name, new_value)

        if isinstance(self, NamedResource) and isinstance(self._name, VarString):
            self._name = ResourceName(self._name.to_string(vars))

    def to_pointer(self):
        return ResourcePointer(
            name=str(self.fqn),
            resource_type=self.resource_type,
        )

    @property
    def container(self):
        return self._container

    @property
    def urn(self):
        return URN.from_resource(self, account_locator="")

    @property
    def fqn(self) -> FQN:
        raise NotImplementedError("Subclasses must implement fqn")


class ResourceContainer:
    def __init__(self):
        self._items: dict[ResourceType, list[Resource]] = {}

    def __contains__(self, item: Resource):
        return item in self._items.get(item.resource_type, [])

    def add(self, *items: Resource):
        if isinstance(items[0], list):
            items = items[0]
        for item in items:
            if not resource_can_be_contained_in(item, self):
                raise WrongContainerException(f"{item} cannot be added to {self}")
            if item.container is not None and not isinstance(item.container, ResourcePointer):
                raise ResourceHasContainerException(f"{item} already belongs to a container")
            item._container = self
            item.requires(self)
            if item.resource_type not in self._items:
                self._items[item.resource_type] = []
            self._items[item.resource_type].append(item)

    def items(self, resource_type: Optional[ResourceType] = None) -> list[Resource]:
        if resource_type:
            return self._items.get(resource_type, [])
        else:
            return list(chain.from_iterable(self._items.values()))

    def find(self, resource_type: ResourceType, name: Union[ResourceName, str]) -> Resource:
        for resource in self.items(resource_type):
            if isinstance(resource, ResourcePointer) and resource.name == name:
                return resource
            elif (
                isinstance(resource, Resource)
                and resource._data is not None
                and getattr(resource._data, "name", None) == name
            ):
                return resource
        raise KeyError(f"Resource {resource_type} {name} not found")

    def remove(self, resource: Resource):
        if resource.resource_type in self._items:
            for index, item in enumerate(self._items[resource.resource_type]):
                if item is resource:
                    self._items[resource.resource_type].pop(index)
                    resource._container = None
            resource.refs = [r for r in resource.refs if r != self]


class NamedResource:
    """
    This class is a mixin that allows resources to be constructed with fully qualified names
    without a bunch of user-land cruft to make it all work. This keeps titan close to the behavior
    of snowflake.

    For example:

    Without this trait, a scoped resource would need to be defined as:
    >>> tbl = Table(name="TBL", database="DB", schema="SCHEMA")

    With this trait, the same table can be defined this way instead:
    >>> tbl = Table(name="DB.SCHEMA.TBL")
    """

    def __init__(self, name: Union[str, ResourceName, VarString], **kwargs):
        if not isinstance(name, (str, ResourceName, VarString)):
            raise TypeError(f"Expected str or ResourceName for name, got {name} ({type(name).__name__}) instead")

        if isinstance(name, (ResourceName, VarString)):
            self._name = name
            super().__init__(**kwargs)
            return

        if string_contains_var(name):
            self._name = VarString(name)
            super().__init__(**kwargs)
            return

        if isinstance(self.scope, AccountScope):
            self._name = ResourceName(name)
            super().__init__(**kwargs)
            return

        try:
            identifier = parse_identifier(name, is_db_scoped=isinstance(self.scope, DatabaseScope))
        except pp.ParseException:
            # Allow identifiers that should be quoted, so long as they aren't insane
            if "." in name:
                raise ValueError(f"Resource name not supported {name}")
            identifier = parse_identifier(f'"{name}"', is_db_scoped=isinstance(self.scope, DatabaseScope))

        self._name = ResourceName(identifier["name"])
        if "database" in identifier and "database" in kwargs:
            raise ValueError("Multiple database names found")
        if "schema" in identifier and "schema" in kwargs:
            raise ValueError("Multiple schema names found")

        super().__init__(
            database=kwargs.pop("database", identifier.get("database")),
            schema=kwargs.pop("schema", identifier.get("schema")),
            **kwargs,
        )

    @property
    def name(self):
        return self._name

    @property
    def fqn(self) -> FQN:
        return self.scope.fully_qualified_name(self.container, self.name)


class ResourcePointer(NamedResource, Resource, ResourceContainer):
    def __init__(self, name: Union[str, ResourceName], resource_type: ResourceType):
        self._resource_type: ResourceType = resource_type
        self.scope = RESOURCE_SCOPES[resource_type]
        super().__init__(name)

        # Don't want to do this for all implicit resources but making an exception for PUBLIC schema
        # If this points to a database, assume it includes a PUBLIC schema
        if self._resource_type == ResourceType.DATABASE and self._name != "SNOWFLAKE":
            self.add(ResourcePointer(name="PUBLIC", resource_type=ResourceType.SCHEMA))
            # self.add(ResourcePointer(name="INFORMATION_SCHEMA", resource_type=ResourceType.SCHEMA))

    def __repr__(self):  # pragma: no cover
        resource_type = getattr(self, "resource_type", None)
        resource_label = resource_label_for_type(resource_type).title() if resource_type else "resource"
        name = getattr(self, "name", None)
        return f"{resource_label}(→{name})"

    def __eq__(self, other):
        if not isinstance(other, ResourcePointer):
            return False
        return self.name == other.name and self.resource_type == other.resource_type

    def __hash__(self):
        return hash((self._name, self._resource_type))

    @property
    def container(self):
        return self._container

    @property
    def fqn(self):
        return self.scope.fully_qualified_name(self.container, self.name)

    @property
    def resource_type(self):
        return self._resource_type

    def to_dict(self, _=None):
        return {
            "_pointer": True,
            "name": self.name,
        }


def convert_to_resource(
    cls: type[Resource],
    resource_or_descriptor: Union[str, dict, Resource, ResourceName],
) -> Resource:
    """
    This function helps provide flexibility to users on how Resource fields are set.
    The most common use case is to allow a user to pass in a string of the name of a resource without
    the need to specify the resource type.

    For example:
    >>> schema = Schema(name="my_schema", owner="some_role")

    Another use case is to allow a user to pass in simple Python data structures for things like table columns.
    >>> table = Table(name='my_table', columns=[{'name': 'id', 'data_type': 'int'}])
    """
    if isinstance(resource_or_descriptor, str) or isinstance(resource_or_descriptor, ResourceName):
        return ResourcePointer(name=resource_or_descriptor, resource_type=cls.resource_type)
    elif isinstance(resource_or_descriptor, dict):
        # We permit two types of anonymous resource descriptors
        # 1. A dict with a single key, the name of the resource
        # Example:
        #   grant = Grant(priv='SELECT', on='table', to={'name': 'some_role'})
        # 2. A dict representing a column or column-like object
        # Example:
        #   table = Table(name='my_table', columns=[{'name': 'id', 'data_type': 'int'}])
        if cls.serialize_inline:
            return cls(**resource_or_descriptor)
        else:
            # This isn't intended to be used by users. This handles the case when a
            # ResourcePointer is deserialized and re-serialized.
            return ResourcePointer(**resource_or_descriptor, resource_type=cls.resource_type)
    elif isinstance(resource_or_descriptor, cls):
        # We are expecting an instance of type cls. If it's a match, return it.
        return resource_or_descriptor
    elif isinstance(resource_or_descriptor, ResourcePointer):
        if resource_or_descriptor.resource_type == cls.resource_type:
            return resource_or_descriptor
        else:
            raise TypeError
    else:
        raise TypeError


def convert_role_ref(role_ref: RoleRef) -> Resource:
    if role_ref.__class__.__name__ == "Role":
        return role_ref
    elif role_ref.__class__.__name__ == "DatabaseRole":
        return role_ref
    elif isinstance(role_ref, ResourcePointer) and role_ref.resource_type in (
        ResourceType.DATABASE_ROLE,
        ResourceType.ROLE,
    ):
        return role_ref
    elif isinstance(role_ref, str) or isinstance(role_ref, ResourceName):
        return ResourcePointer(name=role_ref, resource_type=infer_role_type_from_name(role_ref))
    else:
        raise TypeError


def infer_role_type_from_name(name: Union[str, ResourceName]) -> ResourceType:
    if isinstance(name, ResourceName):
        name = str(name)
    if name == "":
        return ResourceType.ROLE
    identifier = parse_identifier(name, is_db_scoped=True)
    if "database" in identifier:
        return ResourceType.DATABASE_ROLE
    else:
        return ResourceType.ROLE



from dataclasses import dataclass, field
from typing import Optional

from .enums import BlueprintScope, ResourceType, RunMode
from .exceptions import InvalidResourceException, MissingVarException
from .resource_name import ResourceName
from .resources.resource import Resource

_VAR_TYPE_MAP = {
    "bool": bool,
    "boolean": bool,
    "float": float,
    "int": int,
    "integer": int,
    "str": str,
    "string": str,
    "list": list,
}


@dataclass  # (frozen=True)
class BlueprintConfig:
    name: Optional[str] = None
    resources: Optional[list[Resource]] = None
    run_mode: RunMode = RunMode.CREATE_OR_UPDATE
    dry_run: bool = False
    allowlist: Optional[list[ResourceType]] = None
    vars: dict = field(default_factory=dict)
    vars_spec: list[dict] = field(default_factory=list)
    scope: Optional[BlueprintScope] = None
    database: Optional[ResourceName] = None
    schema: Optional[ResourceName] = None

    def __post_init__(self):

        if self.dry_run is None:
            raise ValueError("dry_run must be provided")
        if self.run_mode is None:
            raise ValueError("run_mode must be provided")
        if self.vars is None:
            raise ValueError("vars must be provided")
        if self.vars_spec is None:
            raise ValueError("vars_spec must be provided")

        if not isinstance(self.run_mode, RunMode):
            raise ValueError(f"Invalid run_mode: {self.run_mode}")

        if not isinstance(self.vars, dict):
            raise ValueError(f"vars must be a dictionary, got: {self.vars=}")

        if self.scope is not None and not isinstance(self.scope, BlueprintScope):
            raise ValueError(f"Invalid scope: {self.scope}")

        if self.run_mode == RunMode.SYNC:
            """
            In sync mode, the remote state is not just the resources that were added to the blueprint,
            but all resources that exist in Snowflake. The allowlist is required to limit the scope of
            the sync to a specific set of resources.
            """
            if self.allowlist is None:
                raise ValueError("Sync mode must specify an allowlist")

        if self.allowlist is not None:
            if len(self.allowlist) == 0:
                raise ValueError("Allowlist must have at least one resource type")
            else:
                if self.resources:
                    for resource in self.resources:
                        if resource.resource_type not in self.allowlist:
                            raise InvalidResourceException(
                                f"Resource {resource.urn} of type {resource.resource_type} is not in the allowlist"
                            )

        if self.vars_spec:
            for var in self.vars_spec:
                if var.get("name") is None:
                    raise ValueError("All vars_spec entries must specify a name")
                if var.get("type") is None:
                    raise ValueError("All vars_spec entries must specify a type")
                elif var["type"] not in _VAR_TYPE_MAP:
                    raise ValueError(f"Vars must specify a valid type. Got: {var['type']}")

            # Create a set of all var names in vars_spec for efficient lookup
            spec_names = {var["name"] for var in self.vars_spec}

            # Check each var against its spec
            for var_name, var_value in self.vars.items():
                if var_name not in spec_names:
                    raise ValueError(f"Var '{var_name}' was provided without config")

                spec = next(s for s in self.vars_spec if s["name"] == var_name)
                if not isinstance(var_value, _VAR_TYPE_MAP[spec["type"]]):
                    raise TypeError(f"Var '{var_name}' should be of type {spec['type']}")

            # Check for missing vars and use defaults if available
            # TODO: this causes us to violate frozen=true
            self.vars = set_vars_defaults(self.vars_spec, self.vars)

        if self.scope == BlueprintScope.DATABASE and self.schema is not None:
            raise ValueError("Cannot specify a schema when using DATABASE scope")
        elif self.scope == BlueprintScope.ACCOUNT and (self.database is not None or self.schema is not None):
            raise ValueError(
                f"Cannot specify a database or schema when using ACCOUNT scope (database={repr(self.database)}, schema={repr(self.schema)})"
            )


def set_vars_defaults(vars_spec: list[dict], vars: dict) -> dict:
    new_vars = vars.copy()
    for spec in vars_spec:
        if spec["name"] not in new_vars:
            if "default" in spec:
                new_vars[spec["name"]] = spec["default"]
            else:
                raise MissingVarException(f"Required var '{spec['name']}' is missing and has no default value")
    return new_vars


def print_blueprint_config(config: BlueprintConfig):
    print(f"{config.name=}")
    print(f"config.resources={len(config.resources or [])}")
    print(f"{config.run_mode=}")
    print(f"{config.dry_run=}")
    print(f"{config.allowlist=}")
    print(f"config.vars={list(config.vars.keys())}")



from dataclasses import dataclass

from ..enums import ParseableEnum, ResourceType
from ..props import BoolProp, EnumProp, Props, StringListProp, StringProp
from ..resource_name import ResourceName
from ..scope import AccountScope
from .resource import NamedResource, Resource, ResourceSpec
from .role import Role


class NotificationType(ParseableEnum):
    EMAIL = "EMAIL"
    QUEUE = "QUEUE"


class NotificationProvider(ParseableEnum):
    AWS_SNS = "AWS_SNS"
    GCP_PUBSUB = "GCP_PUBSUB"
    AZURE_STORAGE_QUEUE = "AZURE_STORAGE_QUEUE"
    AZURE_EVENT_GRID = "AZURE_EVENT_GRID"


class NotificationDirection(ParseableEnum):
    INBOUND = "INBOUND"
    OUTBOUND = "OUTBOUND"


@dataclass(unsafe_hash=True)
class _EmailNotificationIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    allowed_recipients: list[str]
    type: NotificationType = NotificationType.EMAIL
    owner: Role = "ACCOUNTADMIN"
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.allowed_recipients is not None and len(self.allowed_recipients) == 0:
            raise ValueError("allowed_recipients can't be empty")


class EmailNotificationIntegration(NamedResource, Resource):
    """
    Description:
        Manages the configuration for email-based notification integrations within Snowflake. This integration
        allows specifying recipients who will receive notifications via email.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-notification-integration

    Fields:
        name (string, required): The name of the email notification integration.
        enabled (bool, required): Specifies whether the notification integration is enabled.
        allowed_recipients (list): A list of email addresses that are allowed to receive notifications.
        comment (string): An optional comment about the notification integration.
        owner (string or Role): The owner role of the notification integration. Defaults to "ACCOUNTADMIN".

    Python:

        ```python
        email_notification_integration = EmailNotificationIntegration(
            name="some_email_notification_integration",
            enabled=True,
            allowed_recipients=["user1@example.com", "user2@example.com"],
            comment="Example email notification integration"
        )
        ```

    Yaml:

        ```yaml
        email_notification_integrations:
          - name: some_email_notification_integration
            enabled: true
            allowed_recipients:
              - user1@example.com
              - user2@example.com
            comment: "Example email notification integration"
        ```
    """

    resource_type = ResourceType.NOTIFICATION_INTEGRATION
    props = Props(
        type=EnumProp("type", [NotificationType.EMAIL]),
        enabled=BoolProp("enabled"),
        allowed_recipients=StringListProp("allowed_recipients", parens=True),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _EmailNotificationIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        allowed_recipients: list[str],
        comment: str = None,
        owner: str = "ACCOUNTADMIN",
        **kwargs,
    ):
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        self._data: _EmailNotificationIntegration = _EmailNotificationIntegration(
            name=self._name,
            enabled=enabled,
            allowed_recipients=allowed_recipients,
            comment=comment,
            owner=owner,
        )


@dataclass(unsafe_hash=True)
class _AWSOutboundNotificationIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    aws_sns_topic_arn: str
    aws_sns_role_arn: str
    direction: NotificationDirection = NotificationDirection.OUTBOUND
    type: str = NotificationType.QUEUE
    notification_provider: str = NotificationProvider.AWS_SNS
    owner: Role = "ACCOUNTADMIN"
    comment: str = None


class AWSOutboundNotificationIntegration(NamedResource, Resource):
    """
    CREATE [ OR REPLACE ] NOTIFICATION INTEGRATION [IF NOT EXISTS]
      <name>
      ENABLED = { TRUE | FALSE }
      DIRECTION = OUTBOUND
      TYPE = QUEUE
      cloudProviderParamsPush
      [ COMMENT = '<string_literal>' ]

    cloudProviderParamsPush (for Amazon SNS) ::=

      NOTIFICATION_PROVIDER = AWS_SNS
      AWS_SNS_TOPIC_ARN = '<topic_arn>'
      AWS_SNS_ROLE_ARN = '<iam_role_arn>'
    """

    resource_type = ResourceType.NOTIFICATION_INTEGRATION
    props = Props(
        enabled=BoolProp("enabled"),
        direction=EnumProp("direction", [NotificationDirection.OUTBOUND]),
        type=EnumProp("type", [NotificationType.QUEUE]),
        notification_provider=EnumProp("notification_provider", [NotificationProvider.AWS_SNS]),
        aws_sns_topic_arn=StringProp("aws_sns_topic_arn"),
        aws_sns_role_arn=StringProp("aws_sns_role_arn"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _AWSOutboundNotificationIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        aws_sns_topic_arn: str,
        aws_sns_role_arn: str,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        kwargs.pop("direction", None)
        kwargs.pop("notification_provider", None)
        super().__init__(name, **kwargs)
        self._data: _AWSOutboundNotificationIntegration = _AWSOutboundNotificationIntegration(
            name=self._name,
            enabled=enabled,
            aws_sns_topic_arn=aws_sns_topic_arn,
            aws_sns_role_arn=aws_sns_role_arn,
            owner=owner,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _GCPOutboundNotificationIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    gcp_pubsub_topic_name: str
    direction: NotificationDirection = NotificationDirection.OUTBOUND
    type: str = NotificationType.QUEUE
    notification_provider: str = NotificationProvider.GCP_PUBSUB
    owner: Role = "ACCOUNTADMIN"
    comment: str = None


class GCPOutboundNotificationIntegration(NamedResource, Resource):
    """
    CREATE [ OR REPLACE ] NOTIFICATION INTEGRATION [IF NOT EXISTS]
      <name>
      ENABLED = { TRUE | FALSE }
      DIRECTION = OUTBOUND
      TYPE = QUEUE
      cloudProviderParamsPush
      [ COMMENT = '<string_literal>' ]

    cloudProviderParamsPush (for Google Pub/Sub) ::=
      NOTIFICATION_PROVIDER = GCP_PUBSUB
      GCP_PUBSUB_TOPIC_NAME = '<topic_id>'
    """

    resource_type = ResourceType.NOTIFICATION_INTEGRATION
    props = Props(
        enabled=BoolProp("enabled"),
        direction=EnumProp("direction", [NotificationDirection.OUTBOUND]),
        type=EnumProp("type", [NotificationType.QUEUE]),
        notification_provider=EnumProp("notification_provider", [NotificationProvider.GCP_PUBSUB]),
        gcp_pubsub_topic_name=StringProp("gcp_pubsub_topic_name"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _GCPOutboundNotificationIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        gcp_pubsub_topic_name: str,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        kwargs.pop("direction", None)
        kwargs.pop("notification_provider", None)
        super().__init__(name, **kwargs)
        self._data: _GCPOutboundNotificationIntegration = _GCPOutboundNotificationIntegration(
            name=self._name,
            enabled=enabled,
            gcp_pubsub_topic_name=gcp_pubsub_topic_name,
            owner=owner,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _AzureOutboundNotificationIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    azure_event_grid_topic_endpoint: str
    azure_tenant_id: str
    direction: NotificationDirection = NotificationDirection.OUTBOUND
    type: str = NotificationType.QUEUE
    notification_provider: str = NotificationProvider.AZURE_EVENT_GRID
    owner: Role = "ACCOUNTADMIN"
    comment: str = None


class AzureOutboundNotificationIntegration(NamedResource, Resource):
    """
    CREATE [ OR REPLACE ] NOTIFICATION INTEGRATION [IF NOT EXISTS]
      <name>
      ENABLED = { TRUE | FALSE }
      DIRECTION = OUTBOUND
      TYPE = QUEUE
      cloudProviderParamsPush
      [ COMMENT = '<string_literal>' ]

    cloudProviderParamsPush (for Microsoft Azure Event Grid) ::=
      NOTIFICATION_PROVIDER = AZURE_EVENT_GRID
      AZURE_EVENT_GRID_TOPIC_ENDPOINT = '<event_grid_topic_endpoint>'
      AZURE_TENANT_ID = '<directory_ID>';
    """

    resource_type = ResourceType.NOTIFICATION_INTEGRATION
    props = Props(
        enabled=BoolProp("enabled"),
        direction=EnumProp("direction", [NotificationDirection.OUTBOUND]),
        type=EnumProp("type", [NotificationType.QUEUE]),
        notification_provider=EnumProp("notification_provider", [NotificationProvider.AZURE_EVENT_GRID]),
        azure_event_grid_topic_endpoint=StringProp("azure_event_grid_topic_endpoint"),
        azure_tenant_id=StringProp("azure_tenant_id"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _AzureOutboundNotificationIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        azure_event_grid_topic_endpoint: str,
        azure_tenant_id: str,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        kwargs.pop("direction", None)
        kwargs.pop("notification_provider", None)
        super().__init__(name, **kwargs)
        self._data: _AzureOutboundNotificationIntegration = _AzureOutboundNotificationIntegration(
            name=self._name,
            enabled=enabled,
            azure_event_grid_topic_endpoint=azure_event_grid_topic_endpoint,
            azure_tenant_id=azure_tenant_id,
            owner=owner,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _GCPInboundNotificationIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    gcp_pubsub_subscription_name: str
    direction: NotificationDirection = NotificationDirection.INBOUND
    type: str = NotificationType.QUEUE
    notification_provider: str = NotificationProvider.GCP_PUBSUB
    owner: Role = "ACCOUNTADMIN"
    comment: str = None


class GCPInboundNotificationIntegration(NamedResource, Resource):
    """
    CREATE [ OR REPLACE ] NOTIFICATION INTEGRATION [IF NOT EXISTS]
      <name>
      ENABLED = { TRUE | FALSE }
      TYPE = QUEUE
      cloudProviderParamsAuto
      [ COMMENT = '<string_literal>' ]

    cloudProviderParamsAuto (for Google Cloud Storage) ::=
      NOTIFICATION_PROVIDER = GCP_PUBSUB
      GCP_PUBSUB_SUBSCRIPTION_NAME = '<subscription_id>'
    """

    resource_type = ResourceType.NOTIFICATION_INTEGRATION
    props = Props(
        enabled=BoolProp("enabled"),
        type=EnumProp("type", [NotificationType.QUEUE]),
        notification_provider=EnumProp("notification_provider", [NotificationProvider.GCP_PUBSUB]),
        gcp_pubsub_subscription_name=StringProp("gcp_pubsub_subscription_name"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _GCPInboundNotificationIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        gcp_pubsub_subscription_name: str,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        kwargs.pop("notification_provider", None)
        super().__init__(name, **kwargs)
        self._data: _GCPInboundNotificationIntegration = _GCPInboundNotificationIntegration(
            name=self._name,
            enabled=enabled,
            gcp_pubsub_subscription_name=gcp_pubsub_subscription_name,
            owner=owner,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _AzureInboundNotificationIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    azure_storage_queue_primary_uri: str
    azure_tenant_id: str
    direction: NotificationDirection = NotificationDirection.INBOUND
    type: str = NotificationType.QUEUE
    notification_provider: str = NotificationProvider.AZURE_STORAGE_QUEUE
    owner: Role = "ACCOUNTADMIN"
    comment: str = None


class AzureInboundNotificationIntegration(NamedResource, Resource):
    """
    CREATE [ OR REPLACE ] NOTIFICATION INTEGRATION [IF NOT EXISTS]
      <name>
      ENABLED = { TRUE | FALSE }
      TYPE = QUEUE
      cloudProviderParamsAuto
      [ COMMENT = '<string_literal>' ]

    cloudProviderParamsAuto (for Microsoft Azure Storage) ::=
      NOTIFICATION_PROVIDER = AZURE_STORAGE_QUEUE
      AZURE_STORAGE_QUEUE_PRIMARY_URI = '<queue_URL>'
      AZURE_TENANT_ID = '<directory_ID>';
    """

    resource_type = ResourceType.NOTIFICATION_INTEGRATION
    props = Props(
        enabled=BoolProp("enabled"),
        type=EnumProp("type", [NotificationType.QUEUE]),
        notification_provider=EnumProp("notification_provider", [NotificationProvider.AZURE_STORAGE_QUEUE]),
        azure_storage_queue_primary_uri=StringProp("azure_storage_queue_primary_uri"),
        azure_tenant_id=StringProp("azure_tenant_id"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _AzureInboundNotificationIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        azure_storage_queue_primary_uri: str,
        azure_tenant_id: str,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        kwargs.pop("notification_provider", None)
        super().__init__(name, **kwargs)
        self._data: _AzureInboundNotificationIntegration = _AzureInboundNotificationIntegration(
            name=self._name,
            enabled=enabled,
            azure_storage_queue_primary_uri=azure_storage_queue_primary_uri,
            azure_tenant_id=azure_tenant_id,
            owner=owner,
            comment=comment,
        )


def _notification_resolver(data: dict):
    if NotificationType(data["type"]) == NotificationType.EMAIL:
        return EmailNotificationIntegration
    if "direction" in data and "notification_provider" in data:
        direction = NotificationDirection(data["direction"])
        provider = NotificationProvider(data["notification_provider"])
        if direction == NotificationDirection.INBOUND:
            if provider == NotificationProvider.AZURE_STORAGE_QUEUE:
                return AzureInboundNotificationIntegration
            elif provider == NotificationProvider.GCP_PUBSUB:
                return GCPInboundNotificationIntegration
        elif direction == NotificationDirection.OUTBOUND:
            if provider == NotificationProvider.AWS_SNS:
                return AWSOutboundNotificationIntegration
            elif provider == NotificationProvider.AZURE_EVENT_GRID:
                return AzureOutboundNotificationIntegration
            elif provider == NotificationProvider.GCP_PUBSUB:
                return GCPOutboundNotificationIntegration
    raise ValueError("Invalid direction or notification provider")


Resource.__resolvers__[ResourceType.NOTIFICATION_INTEGRATION] = _notification_resolver



import copy

import pytest

from dataclasses import fields
from typing import Any, get_args, get_origin

from tests.helpers import get_json_fixtures
from titan.data_types import convert_to_canonical_data_type
from titan.enums import AccountEdition
from titan.resources import Resource
from titan.resource_name import ResourceName
from titan.role_ref import RoleRef

JSON_FIXTURES = list(get_json_fixtures())


def remove_none_values(d):
    new_dict = {}
    for k, v in d.items():
        if isinstance(v, dict):
            new_dict[k] = remove_none_values(v)
        elif isinstance(v, list) and len(v) > 0 and isinstance(v[0], dict):
            new_dict[k] = [remove_none_values(item) for item in v if item is not None]
        elif v is not None:
            new_dict[k] = v
    return new_dict


@pytest.fixture(
    params=JSON_FIXTURES,
    ids=[resource_cls.__name__ for resource_cls, _ in JSON_FIXTURES],
    scope="function",
)
def resource(request):
    resource_cls, data = request.param
    yield resource_cls, data


def _field_type_is_serialized_as_resource_name(field):
    if field.type is Any:
        return False
    if field.type is RoleRef:
        return True
    if field.type is ResourceName:
        return True
    elif isinstance(field.type, str) and field.name == "owner" and field.type == "Role":
        return True
    elif issubclass(field.type, Resource):
        return field.type.serialize_inline is False
    elif get_origin(field.type) is list:
        field_item_type = get_args(field.type)[0]
        if issubclass(field_item_type, Resource):
            return field_item_type.serialize_inline is False

    return False


def _resource_names_are_eq(lhs, rhs):
    if lhs is None and rhs is None:
        return True
    if isinstance(lhs, list):
        return [ResourceName(item) for item in lhs] == [ResourceName(item) for item in rhs]
    else:
        return ResourceName(lhs) == ResourceName(rhs)


def test_data_identity(resource):
    resource_cls, data = resource
    data: dict = copy.deepcopy(data)
    instance = resource_cls(**data)

    serialized: dict = instance.to_dict()
    if "name" in serialized:
        assert _resource_names_are_eq(serialized.pop("name"), data.pop("name"))
    if "columns" in serialized:
        lhs_cols = serialized.pop("columns", []) or []
        rhs_cols = data.pop("columns", []) or []
        assert len(lhs_cols) == len(rhs_cols)
        for lhs, rhs in zip(lhs_cols, rhs_cols):
            if "name" in lhs:
                assert _resource_names_are_eq(lhs.pop("name"), rhs.pop("name"))
            if "data_type" in lhs and "data_type" in rhs:
                assert convert_to_canonical_data_type(lhs.pop("data_type")) == convert_to_canonical_data_type(
                    rhs.pop("data_type")
                )
            assert remove_none_values(lhs) == remove_none_values(rhs)
    if "args" in serialized:
        lhs_args = serialized.pop("args", []) or []
        rhs_args = data.pop("args", []) or []
        assert len(lhs_args) == len(rhs_args)
        for lhs, rhs in zip(lhs_args, rhs_args):
            if "name" in lhs:
                assert _resource_names_are_eq(lhs.pop("name"), rhs.pop("name"))
            assert lhs == rhs
    for field in fields(instance._data):
        if field.name in ["name", "columns", "args"]:
            continue
        if _field_type_is_serialized_as_resource_name(field):
            assert _resource_names_are_eq(serialized.pop(field.name), data.pop(field.name))
        if isinstance(serialized.get(field.name, None), list):
            qqq = serialized.pop(field.name)
            www = data.pop(field.name)
            assert len(qqq) == len(www)
            for lhs, rhs in zip(qqq, www):
                assert lhs == rhs
    assert serialized == data


def test_sql_identity(resource: tuple[type[Resource], dict]):
    resource_cls, data = resource
    if resource_cls.__name__ == "ScannerPackage":
        pytest.skip("Skipping scanner package")
    instance = resource_cls(**data)
    sql = instance.create_sql(AccountEdition.ENTERPRISE)
    new = resource_cls.from_sql(sql)
    new_dict = new.to_dict(AccountEdition.ENTERPRISE)
    instance_dict = instance.to_dict(AccountEdition.ENTERPRISE)
    if "name" in new_dict:
        assert ResourceName(new_dict.pop("name")) == ResourceName(instance_dict.pop("name"))

    assert new_dict == instance_dict



import os
import yaml
import logging
from typing import Any, Optional

from inflection import pluralize
from pathspec import PathSpec
from pathspec.patterns.gitwildmatch import GitWildMatchPattern

from .blueprint_config import BlueprintConfig, set_vars_defaults
from .enums import BlueprintScope, ResourceType, RunMode
from .identifiers import resource_label_for_type, resource_type_for_label
from .resources import (
    Database,
    Resource,
    RoleGrant,
    Schema,
    User,
)
from .resources.resource import ResourcePointer
from .var import string_contains_var, process_for_each

logger = logging.getLogger("titan")

ALIASES = {
    "grants_on_all": ResourceType.GRANT_ON_ALL,
    "account_parameters": ResourceType.ACCOUNT_PARAMETER,
}


def _resources_from_role_grants_config(role_grants_config: list) -> list:
    if len(role_grants_config) == 0:
        return []
    resources = []
    for role_grant in role_grants_config:

        if "to_role" in role_grant:
            resources.append(
                RoleGrant(
                    role=role_grant["role"],
                    to_role=role_grant["to_role"],
                )
            )
        elif "to_user" in role_grant:
            resources.append(
                RoleGrant(
                    role=role_grant["role"],
                    to_user=role_grant["to_user"],
                )
            )
        else:
            for user in role_grant.get("users", []):
                resources.append(
                    RoleGrant(
                        role=role_grant["role"],
                        to_user=user,
                    )
                )
            for to_role in role_grant.get("roles", []):
                resources.append(
                    RoleGrant(
                        role=role_grant["role"],
                        to_role=to_role,
                    )
                )
    if len(resources) == 0:
        raise ValueError(f"No role grants found in config: {role_grants_config}")
    return resources


def _resources_from_database_config(databases_config: list) -> list:
    resources = []
    for database in databases_config:
        schemas = database.pop("schemas", [])
        db = Database(**database)
        resources.append(db)
        for schema in schemas:
            if "owner" not in schema:
                schema["owner"] = db._data.owner
            sch = Schema(**schema)
            db.add(sch)
            resources.append(sch)
    return resources


def _resources_from_users_config(users_config: list) -> list:
    resources = []
    for user in users_config:
        if isinstance(user, dict):
            roles = user.pop("roles", [])
            titan_user = User(**user)
            resources.append(titan_user)
            for role in roles:
                resources.append(
                    RoleGrant(
                        role=role,
                        to_user=titan_user,
                    )
                )
        elif isinstance(user, str):
            resources.append(User.from_sql(user))
    return resources


def process_requires(resource: Resource, requires: list):
    for req in requires:
        resource.requires(ResourcePointer(name=req["name"], resource_type=ResourceType(req["resource_type"])))


def _resources_for_config(config: dict, vars: dict):
    # Special cases
    database_config = config.pop("databases", [])
    role_grants = config.pop("role_grants", [])
    users = config.pop("users", [])

    resources = []
    config_blocks = []

    for resource_type in Resource.__types__.keys():
        resource_label = pluralize(resource_label_for_type(resource_type))
        block = config.pop(resource_label, [])
        if block:
            config_blocks.append((resource_type, block))

    for alias, resource_type in ALIASES.items():
        if alias in config:
            config_blocks.append((resource_type, config.pop(alias)))

    for resource_type, block in config_blocks:
        for resource_data in block:

            if isinstance(resource_data, dict):
                if "for_each" in resource_data:
                    resource_cls = Resource.resolve_resource_cls(resource_type, resource_data)
                    resource_instance = resource_data.copy()
                    for_each = resource_instance.pop("for_each")

                    if isinstance(for_each, str) and for_each.startswith("var."):
                        var_name = for_each.split(".")[1]
                        if var_name not in vars:
                            raise ValueError(f"Var {var_name} not found")
                        for_each_input = vars[var_name]
                    else:
                        raise ValueError(f"for_each must be a var reference. Got: {for_each}")

                    for each_value in for_each_input:
                        for key, value in resource_data.items():
                            if isinstance(value, str) and string_contains_var(value):
                                resource_instance[key] = process_for_each(value, each_value)

                        resource = resource_cls(**resource_instance)
                        resources.append(resource)
                else:
                    requires = resource_data.pop("requires", [])
                    resource_cls = Resource.resolve_resource_cls(resource_type, resource_data)
                    resource = resource_cls(**resource_data)
                    process_requires(resource, requires)
                    resources.append(resource)
            elif isinstance(resource_data, str):
                resource_cls = Resource.resolve_resource_cls(resource_type, {})
                resource = resource_cls.from_sql(resource_data)
                resources.append(resource)
            else:
                raise Exception(f"Unknown resource data type: {resource_data}")

    resources.extend(_resources_from_database_config(database_config))
    resources.extend(_resources_from_role_grants_config(role_grants))
    resources.extend(_resources_from_users_config(users))

    # This code helps resolve grant references to the fully qualified name of the resource.
    # This probably belongs in blueprint as a finalization step.
    # resource_cache = {}
    # for resource in resources:
    #     if hasattr(resource._data, "name"):
    #         resource_cache[(resource.resource_type, ResourceName(resource._data.name))] = resource

    # for resource in resources:
    #     if resource.resource_type == ResourceType.GRANT:
    #         cache_pointer = (resource.on_type, ResourceName(resource.on))
    #         if cache_pointer in resource_cache:
    #             resource._data.on = ResourceName(str(resource_cache[cache_pointer].fqn))

    return resources


def collect_blueprint_config(yaml_config: dict, cli_config: Optional[dict[str, Any]] = None) -> BlueprintConfig:
    yaml_config_ = yaml_config.copy()
    cli_config_ = cli_config.copy() if cli_config else {}
    blueprint_args: dict[str, Any] = {}

    for key in ["allowlist", "dry_run", "name", "run_mode"]:
        if key in yaml_config_ and key in cli_config_:
            raise ValueError(f"Cannot specify `{key}` in both yaml config and cli")

    allowlist = yaml_config_.pop("allowlist", None) or cli_config_.pop("allowlist", None)
    database = yaml_config_.pop("database", None) or cli_config_.pop("database", None)
    dry_run = yaml_config_.pop("dry_run", None) or cli_config_.pop("dry_run", None)
    name = yaml_config_.pop("name", None) or cli_config_.pop("name", None)
    run_mode = yaml_config_.pop("run_mode", None) or cli_config_.pop("run_mode", None)
    scope = yaml_config_.pop("scope", None) or cli_config_.pop("scope", None)
    schema = yaml_config_.pop("schema", None) or cli_config_.pop("schema", None)
    input_vars = cli_config_.pop("vars", {}) or {}
    vars_spec = yaml_config_.pop("vars", [])

    if allowlist:
        blueprint_args["allowlist"] = [ResourceType(resource_type) for resource_type in allowlist]

    if database:
        blueprint_args["database"] = database

    if dry_run:
        blueprint_args["dry_run"] = dry_run

    if name:
        blueprint_args["name"] = name

    if run_mode:
        blueprint_args["run_mode"] = RunMode(run_mode)

    if scope:
        blueprint_args["scope"] = BlueprintScope(scope)

    if schema:
        blueprint_args["schema"] = schema

    blueprint_args["vars"] = input_vars

    if vars_spec:
        if not isinstance(vars_spec, list):
            raise ValueError("vars config entry must be a list of dicts")
        blueprint_args["vars_spec"] = vars_spec
        blueprint_args["vars"] = set_vars_defaults(vars_spec, blueprint_args["vars"])

    resources = _resources_for_config(yaml_config_, blueprint_args["vars"])

    if len(resources) == 0:
        raise ValueError("No resources found in config")

    blueprint_args["resources"] = resources

    if yaml_config_:
        raise ValueError(f"Unknown keys in config: {yaml_config_.keys()}")

    return BlueprintConfig(**blueprint_args)


def crawl(path: str):
    # Load .titanignore patterns if the file exists
    gitignore_path = os.path.join(path, ".titanignore")
    if os.path.exists(gitignore_path):
        with open(gitignore_path) as f:
            spec = PathSpec.from_lines(GitWildMatchPattern, f.readlines())
    else:
        spec = PathSpec([])

    if os.path.isfile(path):
        yield path
        return

    for root, _, files in os.walk(path):
        for file in files:
            if file.endswith(".yaml") or file.endswith(".yml"):
                full_path = os.path.join(root, file)
                # Get path relative to the base path for titanignore matching
                rel_path = os.path.relpath(full_path, path)
                if not spec.match_file(rel_path):
                    yield full_path


def read_config(config_path) -> dict:
    with open(config_path, "r") as f:
        try:
            config = yaml.safe_load(f)
        except yaml.YAMLError as e:
            raise ValueError(f"Error parsing YAML file: {config_path}") from e
    return config


def merge_configs(config1: dict, config2: dict) -> dict:
    merged = config1.copy()
    for key, value in config2.items():
        if key in merged:
            if isinstance(merged[key], list):
                merged[key] = merged[key] + value
            elif merged[key] is None:
                merged[key] = value
            else:
                raise ValueError(f"Found a conflict for key `{key}` with {value} and {merged[key]}")
        else:
            merged[key] = value
    return merged


def collect_configs_from_path(path: str) -> list[tuple[str, dict]]:
    configs = []

    if not os.path.exists(path):
        raise ValueError(f"Invalid path: `{path}`. Must be a file or directory.")

    for file in crawl(path):
        config = read_config(file)
        configs.append((file, config))

    if len(configs) == 0:
        raise ValueError(f"No valid YAML files were read from the given path: {path}")

    return configs


def parse_resources(resource_labels_str: Optional[str]) -> Optional[list[ResourceType]]:
    if resource_labels_str is None or resource_labels_str == "all":
        return None
    return [resource_type_for_label(resource_label) for resource_label in resource_labels_str.split(",")]


def collect_vars_from_environment() -> dict:
    vars = {}
    for key, value in os.environ.items():
        if key.startswith("TITAN_VAR_"):
            vars[key[10:].lower()] = value
    return vars


def merge_vars(vars: dict, other_vars: dict) -> dict:
    for key in other_vars.keys():
        if key in vars:
            raise ValueError(f"Conflicting var found: '{key}'")
    merged = vars.copy()
    merged.update(other_vars)
    return merged



import pytest
from inflection import pluralize

from tests.helpers import get_json_fixtures
from titan.gitops import collect_blueprint_config
from titan.identifiers import resource_label_for_type

JSON_FIXTURES = list(get_json_fixtures())


@pytest.fixture
def database_config() -> dict:
    return {
        "databases": [
            {
                "name": "test_database",
                "comment": "test database",
                "schemas": [
                    {
                        "name": "test_schema",
                        "comment": "test schema",
                    }
                ],
            }
        ]
    }


@pytest.fixture
def resource_config() -> dict:
    config = {}
    for resource_cls, resource_config in JSON_FIXTURES:
        config[pluralize(resource_label_for_type(resource_cls.resource_type))] = [resource_config]

    return config


def test_database_config(database_config):
    blueprint_config = collect_blueprint_config(database_config)
    assert len(blueprint_config.resources) == 2


@pytest.mark.skip(reason="JSON_FIXTURES doesn't include things like role grants yet")
def test_resource_config(resource_config):
    resources = collect_blueprint_config(resource_config)
    resource_types = set([resource.resource_type for resource in resources])
    expected_resource_types = set([resource_cls.resource_type for resource_cls, _ in JSON_FIXTURES])
    assert resource_types == expected_resource_types


def test_grant_on_all_alias():
    config_base = {
        "grant_on_alls": [
            {
                "priv": "SELECT",
                "on_all_tables_in_schema": "sch",
                "to": "somerole",
            }
        ]
    }
    config_aliased = {
        "grants_on_all": [
            {
                "priv": "SELECT",
                "on_all_tables_in_schema": "sch",
                "to": "somerole",
            }
        ]
    }
    blueprint_config = collect_blueprint_config(config_base)
    blueprint_config_aliased = collect_blueprint_config(config_aliased)
    assert len(blueprint_config.resources) == 1
    assert len(blueprint_config_aliased.resources) == 1
    assert blueprint_config.resources[0]._data == blueprint_config_aliased.resources[0]._data


def test_vars_type_validation(database_config):
    yaml_config = {
        "vars": [{"name": "foo", "type": "string"}],
        **database_config,
    }
    cli_config = {
        "vars": {"foo": 42},
    }
    with pytest.raises(TypeError):
        collect_blueprint_config(yaml_config, cli_config)

    yaml_config = {
        "vars": [{"name": "foo", "type": "int", "default": 0}],
        **database_config,
    }
    cli_config = {
        "vars": {"foo": "bar"},
    }
    with pytest.raises(TypeError):
        collect_blueprint_config(yaml_config, cli_config)


def test_vars_defaults(database_config):
    config = {
        "vars": [{"name": "foo", "default": "bar", "type": "string"}],
        **database_config,
    }
    blueprint_config = collect_blueprint_config(config)
    assert blueprint_config.vars["foo"] == "bar"


def test_for_each():
    config = {
        "vars": [{"name": "some_list_var", "default": ["bar", "baz"], "type": "list"}],
        "roles": [
            {
                "for_each": "var.some_list_var",
                "name": "role_{{ each.value}}",
            }
        ],
    }
    blueprint_config = collect_blueprint_config(config)
    assert blueprint_config.resources is not None
    assert len(blueprint_config.resources) == 2
    assert [resource.urn.fqn.name for resource in blueprint_config.resources] == ["role_bar", "role_baz"]



import json
from copy import deepcopy

import pytest

from titan import resources as res
from titan import var
from titan.blueprint import (
    Blueprint,
    CreateResource,
    _merge_pointers,
    compile_plan_to_sql,
    dump_plan,
)
from titan.blueprint_config import BlueprintConfig
from titan.enums import AccountEdition, BlueprintScope, ResourceType, RunMode
from titan.exceptions import (
    DuplicateResourceException,
    InvalidResourceException,
    MissingVarException,
    NonConformingPlanException,
    WrongEditionException,
)
from titan.identifiers import FQN, URN, parse_URN
from titan.resource_name import ResourceName
from titan.resources.resource import ResourcePointer
from titan.var import VarString


@pytest.fixture
def session_ctx() -> dict:
    return {
        "account": "SOMEACCT",
        "account_edition": AccountEdition.ENTERPRISE,
        "account_locator": "ABCD123",
        "role": "SYSADMIN",
        "available_roles": [
            "SYSADMIN",
            "USERADMIN",
            "ACCOUNTADMIN",
            "SECURITYADMIN",
            "PUBLIC",
        ],
    }


@pytest.fixture
def remote_state() -> dict:
    return {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
    }


@pytest.fixture
def resource_manifest():
    session_ctx = {
        "account": "SOMEACCT",
        "account_edition": AccountEdition.ENTERPRISE,
        "account_locator": "ABCD123",
        "current_role": "SYSADMIN",
        "available_roles": ["SYSADMIN", "USERADMIN"],
    }
    db = res.Database(name="DB")
    schema = res.Schema(name="SCHEMA", database=db)
    table = res.Table(name="TABLE", columns=[{"name": "ID", "data_type": "INT"}])
    schema.add(table)
    view = res.View(name="VIEW", schema=schema, as_="SELECT 1")
    udf = res.PythonUDF(
        name="SOMEUDF",
        returns="VARCHAR",
        args=[],
        runtime_version="3.9",
        handler="main",
        comment="This is a UDF comment",
    )
    schema.add(udf)
    blueprint = Blueprint(name="blueprint", resources=[db, table, schema, view, udf])
    manifest = blueprint.generate_manifest(session_ctx)
    return manifest


def test_blueprint_with_database(resource_manifest):

    db_urn = parse_URN("urn::ABCD123:database/DB")
    assert db_urn in resource_manifest
    assert resource_manifest[db_urn].data == {
        "name": "DB",
        "owner": "SYSADMIN",
        "comment": None,
        "catalog": None,
        "external_volume": None,
        "data_retention_time_in_days": 1,
        "default_ddl_collation": None,
        "max_data_extension_time_in_days": 14,
        "transient": False,
    }


def test_blueprint_with_schema(resource_manifest):
    schema_urn = parse_URN("urn::ABCD123:schema/DB.SCHEMA")
    assert schema_urn in resource_manifest
    assert resource_manifest[schema_urn].data == {
        "comment": None,
        "data_retention_time_in_days": 1,
        "default_ddl_collation": None,
        "managed_access": False,
        "max_data_extension_time_in_days": 14,
        "name": "SCHEMA",
        "owner": "SYSADMIN",
        "transient": False,
    }


def test_blueprint_with_view(resource_manifest):
    view_urn = parse_URN("urn::ABCD123:view/DB.SCHEMA.VIEW")
    assert view_urn in resource_manifest
    assert resource_manifest[view_urn].data == {
        "as_": "SELECT 1",
        "change_tracking": False,
        "columns": None,
        "comment": None,
        "copy_grants": False,
        "name": "VIEW",
        "owner": "SYSADMIN",
        "recursive": None,
        "secure": False,
        "volatile": None,
    }


def test_blueprint_with_table(resource_manifest):
    table_urn = parse_URN("urn::ABCD123:table/DB.SCHEMA.TABLE")
    assert table_urn in resource_manifest
    assert resource_manifest[table_urn].data == {
        "name": "TABLE",
        "owner": "SYSADMIN",
        "columns": [
            {
                "name": "ID",
                "data_type": "NUMBER(38,0)",
                "collate": None,
                "comment": None,
                "constraint": None,
                "not_null": False,
                "default": None,
                "tags": None,
            }
        ],
        "constraints": None,
        "transient": False,
        "cluster_by": None,
        "enable_schema_evolution": False,
        "data_retention_time_in_days": None,
        "max_data_extension_time_in_days": None,
        "change_tracking": False,
        "default_ddl_collation": None,
        "copy_grants": None,
        "row_access_policy": None,
        "comment": None,
    }


def test_blueprint_with_udf(resource_manifest):
    # parse URN is incorrectly stripping the parens. Not sure what the correct behavior should be
    # udf_urn = parse_URN("urn::ABCD123:function/DB.PUBLIC.SOMEUDF()")
    udf_urn = URN(
        resource_type=ResourceType.FUNCTION,
        fqn=FQN(
            database=ResourceName("DB"),
            schema=ResourceName("SCHEMA"),
            name=ResourceName("SOMEUDF"),
            arg_types=[],
        ),
        account_locator="ABCD123",
    )
    assert udf_urn in resource_manifest
    assert resource_manifest[udf_urn].data == {
        "name": "SOMEUDF",
        "owner": "SYSADMIN",
        "returns": "VARCHAR",
        "handler": "main",
        "runtime_version": "3.9",
        "comment": "This is a UDF comment",
        "args": [],
        "as_": None,
        "copy_grants": False,
        "language": "PYTHON",
        "external_access_integrations": None,
        "imports": None,
        "null_handling": None,
        "packages": None,
        "secrets": None,
        "secure": None,
        "volatility": None,
    }


def test_blueprint_resource_owned_by_plan_role(session_ctx, remote_state):
    role = res.Role("SOME_ROLE")
    wh = res.Warehouse("WH", owner=role)
    grant = res.RoleGrant(role=role, to_role="SYSADMIN")
    blueprint = Blueprint(name="blueprint", resources=[wh, role, grant])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)

    plan_urns = [change.urn for change in plan]
    assert plan_urns == [
        parse_URN("urn::ABCD123:role/SOME_ROLE"),
        parse_URN("urn::ABCD123:role_grant/SOME_ROLE?role=SYSADMIN"),
        parse_URN("urn::ABCD123:warehouse/WH"),
    ]

    changes = compile_plan_to_sql(session_ctx, plan)
    assert len(changes) == 8
    assert changes[0] == "USE SECONDARY ROLES ALL"
    assert changes[1] == "USE ROLE USERADMIN"
    assert changes[2] == "CREATE ROLE SOME_ROLE"
    assert changes[3] == "USE ROLE SECURITYADMIN"
    assert changes[4] == "GRANT ROLE SOME_ROLE TO ROLE SYSADMIN"
    assert changes[5] == f"USE ROLE {session_ctx['role']}"
    assert changes[6].startswith("CREATE WAREHOUSE WH")
    assert changes[7] == "GRANT OWNERSHIP ON WAREHOUSE WH TO ROLE SOME_ROLE COPY CURRENT GRANTS"


def test_blueprint_deduplicate_resources(session_ctx, remote_state):
    blueprint = Blueprint(
        name="blueprint",
        resources=[
            res.Database("DB"),
            ResourcePointer(name="DB", resource_type=ResourceType.DATABASE),
        ],
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:database/DB")
    assert plan[0].resource_cls == res.Database

    blueprint = Blueprint(
        name="blueprint",
        resources=[
            res.Database("DB"),
            res.Database("DB", comment="This is a comment"),
        ],
    )
    with pytest.raises(DuplicateResourceException):
        blueprint.generate_manifest(session_ctx)

    blueprint = Blueprint(
        name="blueprint",
        resources=[
            res.Grant(priv="USAGE", on_database="DB", to="SOME_ROLE"),
            res.Grant(priv="USAGE", on_database="DB", to="SOME_ROLE"),
        ],
    )
    with pytest.raises(DuplicateResourceException):
        blueprint.generate_manifest(session_ctx)


def test_blueprint_dont_add_public_schema(session_ctx, remote_state):
    db = res.Database("DB")
    public = ResourcePointer(name="PUBLIC", resource_type=ResourceType.SCHEMA)
    blueprint = Blueprint(
        name="blueprint",
        resources=[db, public],
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:database/DB")
    assert plan[0].resource_cls == res.Database


def test_blueprint_implied_container_tree(session_ctx, remote_state):
    remote_state[parse_URN("urn::ABCD123:database/STATIC_DB")] = {"owner": "SYSADMIN"}
    remote_state[parse_URN("urn::ABCD123:schema/STATIC_DB.PUBLIC")] = {"owner": "SYSADMIN"}
    func = res.JavascriptUDF(
        name="func", args=[], returns="INT", as_="return 1;", database="STATIC_DB", schema="public"
    )
    blueprint = Blueprint(name="blueprint", resources=[func])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn.fqn.name == "FUNC"
    assert plan[0].resource_cls == res.JavascriptUDF


def test_blueprint_chained_ownership(session_ctx, remote_state):
    role = res.Role("SOME_ROLE")
    role_grant = res.RoleGrant(role=role, to_role="SYSADMIN")
    db = res.Database("DB", owner=role)
    schema = res.Schema("SCHEMA", database=db, owner=role)
    blueprint = Blueprint(name="blueprint", resources=[db, schema, role_grant, role])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 4
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:role/SOME_ROLE")
    assert plan[0].resource_cls == res.Role
    assert isinstance(plan[1], CreateResource)
    assert plan[1].urn == parse_URN("urn::ABCD123:role_grant/SOME_ROLE?role=SYSADMIN")
    assert plan[1].resource_cls == res.RoleGrant
    assert isinstance(plan[2], CreateResource)
    assert plan[2].urn == parse_URN("urn::ABCD123:database/DB")
    assert plan[2].resource_cls == res.Database
    assert isinstance(plan[3], CreateResource)
    assert plan[3].urn == parse_URN("urn::ABCD123:schema/DB.SCHEMA")
    assert plan[3].resource_cls == res.Schema


def test_blueprint_polymorphic_resource_resolution(session_ctx, remote_state):

    role = res.Role(name="DEMO_ROLE")
    sysad_grant = res.RoleGrant(role=role, to_role="SYSADMIN")
    test_db = res.Database(name="TEST_TITAN", transient=False, data_retention_time_in_days=1, comment="Test Titan")
    schema = res.Schema(name="TEST_SCHEMA", database=test_db, transient=False, comment="Test Titan Schema")
    warehouse = res.Warehouse(name="FAKER_LOADER", auto_suspend=60)

    future_schema_grant = res.FutureGrant(priv="usage", on_future_schemas_in=test_db, to=role)
    post_grant = [future_schema_grant]

    grants = [
        res.Grant(priv="usage", to=role, on=warehouse),
        res.Grant(priv="operate", to=role, on=warehouse),
        res.Grant(priv="usage", to=role, on=test_db),
        # future_schema_grant,
        # x
        # Grant(priv="usage", to=role, on=schema)
    ]

    sales_table = res.Table(
        name="faker_data",
        schema=schema,
        columns=[
            res.Column(name="NAME", data_type="VARCHAR(16777216)"),
            res.Column(name="EMAIL", data_type="VARCHAR(16777216)"),
            res.Column(name="ADDRESS", data_type="VARCHAR(16777216)"),
            res.Column(name="ORDERED_AT_UTC", data_type="NUMBER(38,0)"),
            res.Column(name="EXTRACTED_AT_UTC", data_type="NUMBER(38,0)"),
            res.Column(name="SALES_ORDER_ID", data_type="VARCHAR(16777216)"),
        ],
        comment="Test Table",
    )
    blueprint = Blueprint(
        name="blueprint",
        resources=[
            role,
            sysad_grant,
            # user_grant,
            test_db,
            # *pre_grant,
            schema,
            sales_table,
            # pipe,
            warehouse,
            *grants,
        ],
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 9


def test_blueprint_scope_sorting(session_ctx, remote_state):
    db = res.Database(name="DB")
    schema = res.Schema(name="SCHEMA", database=db)
    view = res.View(name="SOME_VIEW", schema=schema, as_="SELECT 1")
    blueprint = Blueprint(name="blueprint", resources=[view, schema, db])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 3
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:database/DB")
    assert plan[0].resource_cls == res.Database
    assert isinstance(plan[1], CreateResource)
    assert plan[1].urn == parse_URN("urn::ABCD123:schema/DB.SCHEMA")
    assert plan[1].resource_cls == res.Schema
    assert isinstance(plan[2], CreateResource)
    assert plan[2].urn == parse_URN("urn::ABCD123:view/DB.SCHEMA.SOME_VIEW")
    assert plan[2].resource_cls == res.View


def test_blueprint_reference_sorting(session_ctx, remote_state):
    db1 = res.Database(name="DB1")
    db2 = res.Database(name="DB2")
    db2.requires(db1)
    db3 = res.Database(name="DB3")
    db3.requires(db2)
    blueprint = Blueprint(resources=[db3, db1, db2])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 3
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:database/DB1")
    assert plan[0].resource_cls == res.Database
    assert isinstance(plan[1], CreateResource)
    assert plan[1].urn == parse_URN("urn::ABCD123:database/DB2")
    assert plan[1].resource_cls == res.Database
    assert isinstance(plan[2], CreateResource)
    assert plan[2].urn == parse_URN("urn::ABCD123:database/DB3")
    assert plan[2].resource_cls == res.Database


def test_blueprint_ownership_sorting(session_ctx, remote_state):

    role = res.Role(name="SOME_ROLE")
    role_grant = res.RoleGrant(role=role, to_role="SYSADMIN")
    wh = res.Warehouse(name="WH", owner=role)

    blueprint = Blueprint(resources=[wh, role_grant, role])
    manifest = blueprint.generate_manifest(session_ctx)

    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 3
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:role/SOME_ROLE")
    assert plan[0].resource_cls == res.Role
    assert isinstance(plan[1], CreateResource)
    assert plan[1].urn == parse_URN("urn::ABCD123:role_grant/SOME_ROLE?role=SYSADMIN")
    assert plan[1].resource_cls == res.RoleGrant
    assert isinstance(plan[2], CreateResource)
    assert plan[2].urn == parse_URN("urn::ABCD123:warehouse/WH")
    assert plan[2].resource_cls == res.Warehouse

    sql = compile_plan_to_sql(session_ctx, plan)
    assert len(sql) == 8
    assert sql[0] == "USE SECONDARY ROLES ALL"
    assert sql[1] == "USE ROLE USERADMIN"
    assert sql[2] == "CREATE ROLE SOME_ROLE"
    assert sql[3] == "USE ROLE SECURITYADMIN"
    assert sql[4] == "GRANT ROLE SOME_ROLE TO ROLE SYSADMIN"
    assert sql[5] == f"USE ROLE {session_ctx['role']}"
    assert sql[6].startswith("CREATE WAREHOUSE WH")
    assert sql[7] == "GRANT OWNERSHIP ON WAREHOUSE WH TO ROLE SOME_ROLE COPY CURRENT GRANTS"


def test_blueprint_dump_plan_create(session_ctx, remote_state):
    blueprint = Blueprint(resources=[res.Role("role1")])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    plan_json_str = dump_plan(plan, format="json")
    assert json.loads(plan_json_str) == [
        {
            "action": "CREATE",
            "resource_cls": "Role",
            "urn": "urn::ABCD123:role/ROLE1",
            "after": {"name": "ROLE1", "owner": "USERADMIN", "comment": None},
        }
    ]
    plan_str = dump_plan(plan, format="text")
    assert (
        plan_str
        == """
» titan core
» Plan: 1 to create, 0 to update, 0 to transfer, 0 to drop.

+ urn::ABCD123:role/ROLE1 {
  + name    = "ROLE1"
  + owner   = "USERADMIN"
  + comment = None
}

"""
    )


def test_blueprint_dump_plan_update(session_ctx):
    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:role/ROLE1"): {
            "name": "ROLE1",
            "owner": "USERADMIN",
            "comment": "old",
        },
    }
    blueprint = Blueprint(resources=[res.Role("role1", comment="new")])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    plan_json_str = dump_plan(plan, format="json")
    assert json.loads(plan_json_str) == [
        {
            "action": "UPDATE",
            "resource_cls": "Role",
            "urn": "urn::ABCD123:role/ROLE1",
            "before": {"name": "ROLE1", "owner": "USERADMIN", "comment": "old"},
            "after": {"name": "ROLE1", "owner": "USERADMIN", "comment": "new"},
            "delta": {"comment": "new"},
        }
    ]
    plan_str = dump_plan(plan, format="text")
    assert (
        plan_str
        == """
» titan core
» Plan: 0 to create, 1 to update, 0 to transfer, 0 to drop.

~ urn::ABCD123:role/ROLE1 {
  ~ comment = "old" -> "new"
}

"""
    )


def test_blueprint_dump_plan_transfer(session_ctx):
    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:role/ROLE1"): {
            "name": "ROLE1",
            "owner": "ACCOUNTADMIN",
            "comment": None,
        },
    }
    blueprint = Blueprint(resources=[res.Role("role1", owner="USERADMIN")])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    plan_json_str = dump_plan(plan, format="json")
    assert json.loads(plan_json_str) == [
        {
            "action": "TRANSFER",
            "resource_cls": "Role",
            "urn": "urn::ABCD123:role/ROLE1",
            "from_owner": "ACCOUNTADMIN",
            "to_owner": "USERADMIN",
        }
    ]
    plan_str = dump_plan(plan, format="text")
    assert (
        plan_str
        == """
» titan core
» Plan: 0 to create, 0 to update, 1 to transfer, 0 to drop.

~ urn::ABCD123:role/ROLE1 {
  ~ owner = "ACCOUNTADMIN" -> "USERADMIN"
}

"""
    )


def test_blueprint_dump_plan_drop(session_ctx):
    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:role/ROLE1"): {
            "name": "ROLE1",
            "owner": "ACCOUNTADMIN",
            "comment": None,
        },
    }
    blueprint = Blueprint(resources=[], run_mode="SYNC", allowlist=[ResourceType.ROLE])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    plan_json_str = dump_plan(plan, format="json")
    plan_dict = json.loads(plan_json_str)
    assert len(plan_dict) == 1
    assert plan_dict[0] == {
        "action": "DROP",
        "urn": "urn::ABCD123:role/ROLE1",
        "before": {"name": "ROLE1", "owner": "ACCOUNTADMIN", "comment": None},
    }

    plan_str = dump_plan(plan, format="text")
    assert (
        plan_str
        == """
» titan core
» Plan: 0 to create, 0 to update, 0 to transfer, 1 to drop.

- urn::ABCD123:role/ROLE1

"""
    )


def test_blueprint_vars(session_ctx):
    blueprint = Blueprint(
        resources=[res.Role(name="role", comment=var.role_comment)],
        vars={"role_comment": "var role comment"},
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert manifest.resources[1].data["comment"] == "var role comment"

    role = res.Role(name="role", comment="some comment {{ var.suffix }}")
    assert isinstance(role._data.comment, VarString)
    blueprint = Blueprint(
        resources=[role],
        vars={"suffix": "1234"},
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert manifest.resources[1].data["comment"] == "some comment 1234"

    role = res.Role(name=var.role_name)
    assert isinstance(role.name, VarString)
    blueprint = Blueprint(
        resources=[role],
        vars={"role_name": "role123"},
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert manifest.resources[1].data["name"] == "role123"

    role = res.Role(name="role_{{ var.suffix }}")
    assert isinstance(role.name, VarString)
    blueprint = Blueprint(
        resources=[role],
        vars={"suffix": "5678"},
    )
    manifest = blueprint.generate_manifest(session_ctx)
    assert manifest.resources[1].data["name"] == "role_5678"


def test_blueprint_vars_spec(session_ctx):
    blueprint = Blueprint(
        resources=[res.Role(name="role", comment=var.role_comment)],
        vars_spec=[
            {
                "name": "role_comment",
                "type": "string",
                "default": "var role comment",
            }
        ],
    )
    assert blueprint._config.vars == {"role_comment": "var role comment"}
    manifest = blueprint.generate_manifest(session_ctx)
    assert manifest.resources[1].data["comment"] == "var role comment"

    with pytest.raises(MissingVarException):
        blueprint = Blueprint(
            resources=[res.Role(name="role", comment=var.role_comment)],
            vars_spec=[{"name": "role_comment", "type": "string"}],
        )

    blueprint = Blueprint(resources=[res.Role(name="role", comment=var.role_comment)])
    with pytest.raises(MissingVarException):
        blueprint.generate_manifest(session_ctx)


def test_blueprint_allowlist(session_ctx, remote_state):
    blueprint = Blueprint(
        resources=[res.Role(name="role1")],
        allowlist=[ResourceType.ROLE],
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1

    blueprint = Blueprint(allowlist=["ROLE"])
    assert blueprint._config.allowlist == [ResourceType.ROLE]
    with pytest.raises(InvalidResourceException):
        blueprint.add(res.Database(name="db1"))

    with pytest.raises(InvalidResourceException):
        blueprint = Blueprint(
            resources=[res.Role(name="role1")],
            allowlist=[ResourceType.DATABASE],
        )


def test_blueprint_config_validation():
    with pytest.raises(ValueError):
        BlueprintConfig(run_mode=None)
    with pytest.raises(ValueError):
        BlueprintConfig(run_mode="non-existent-mode")
    with pytest.raises(ValueError):
        BlueprintConfig(run_mode="sync")
    with pytest.raises(ValueError):
        BlueprintConfig(allowlist=[])

    bp = Blueprint(run_mode="SYNC", allowlist=["ROLE"])
    assert bp._config.run_mode == RunMode.SYNC
    bp = Blueprint(run_mode="CREATE-OR-UPDATE")
    assert bp._config.run_mode == RunMode.CREATE_OR_UPDATE

    bp = Blueprint(allowlist=["ROLE"])
    assert bp._config.allowlist == [ResourceType.ROLE]

    with pytest.raises(ValueError):
        Blueprint(allowlist=["non-existent-resource-type"])


def test_merge_account_scoped_resources():
    resources = [
        res.Database(name="DB1"),
        ResourcePointer(name="DB1", resource_type=ResourceType.DATABASE),
    ]
    merged = _merge_pointers(resources)
    assert len(merged) == 1
    assert isinstance(merged[0], res.Database)
    assert merged[0].name == "DB1"

    resources = [
        res.Database(name="DB1"),
        res.Database(name="DB2"),
    ]
    merged = _merge_pointers(resources)
    assert len(merged) == 2


def test_merge_account_scoped_resources_fail():
    resources = [
        res.Database(name="DB1"),
        res.Database(name="DB1", comment="namespace conflict"),
    ]
    with pytest.raises(DuplicateResourceException):
        _merge_pointers(resources)


def test_blueprint_edition_checks(session_ctx, remote_state):
    session_ctx = deepcopy(session_ctx)
    session_ctx["account_edition"] = AccountEdition.STANDARD

    blueprint = Blueprint(resources=[res.Database(name="DB1"), res.Tag(name="TAG1")])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    with pytest.raises(NonConformingPlanException):
        blueprint._raise_for_nonconforming_plan(session_ctx, plan)

    blueprint = Blueprint(resources=[res.Warehouse(name="WH", min_cluster_count=2)])
    with pytest.raises(WrongEditionException):
        blueprint.generate_manifest(session_ctx)

    blueprint = Blueprint(resources=[res.Warehouse(name="WH", min_cluster_count=1)])
    assert blueprint.generate_manifest(session_ctx)

    blueprint = Blueprint(resources=[res.Warehouse(name="WH")])
    assert blueprint.generate_manifest(session_ctx)


def test_blueprint_warehouse_scaling_policy_doesnt_render_in_standard_edition(session_ctx, remote_state):
    session_ctx = deepcopy(session_ctx)
    session_ctx["account_edition"] = AccountEdition.STANDARD
    wh = res.Warehouse(name="WH", warehouse_size="XSMALL")
    blueprint = Blueprint(resources=[wh])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    sql = compile_plan_to_sql(session_ctx, plan)
    assert len(sql) == 3
    assert sql[0] == "USE SECONDARY ROLES ALL"
    assert sql[1] == "USE ROLE SYSADMIN"
    assert sql[2].startswith("CREATE WAREHOUSE WH")
    assert "scaling_policy" not in sql[2]


def test_blueprint_scope_config():

    bc = BlueprintConfig(
        scope=BlueprintScope.DATABASE,
        database=ResourceName("foo"),
    )
    assert bc

    with pytest.raises(ValueError):
        BlueprintConfig(
            scope=BlueprintScope.DATABASE,
            schema=ResourceName("bar"),
        )

    with pytest.raises(ValueError):
        BlueprintConfig(
            scope=BlueprintScope.ACCOUNT,
            database=ResourceName("foo"),
        )

    with pytest.raises(ValueError):
        BlueprintConfig(
            scope=BlueprintScope.ACCOUNT,
            schema=ResourceName("bar"),
        )

    with pytest.raises(ValueError):
        BlueprintConfig(
            scope=BlueprintScope.ACCOUNT,
            database=ResourceName("foo"),
            schema=ResourceName("bar"),
        )


def test_blueprint_scope(session_ctx, remote_state):

    blueprint = Blueprint(resources=[res.Database(name="DB1")], scope=BlueprintScope.DATABASE)
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1

    blueprint = Blueprint(resources=[res.Role(name="ROLE1")], scope=BlueprintScope.DATABASE)
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    with pytest.raises(NonConformingPlanException):
        blueprint._raise_for_nonconforming_plan(session_ctx, plan)

    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:database/DB1"): {"owner": "SYSADMIN"},
        parse_URN("urn::ABCD123:schema/DB1.PUBLIC"): {"owner": "SYSADMIN"},
    }

    blueprint = Blueprint(
        resources=[
            res.Schema(name="SCHEMA1"),
            res.Task(name="TASK1"),
        ],
        scope=BlueprintScope.DATABASE,
        database="DB1",
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 2

    blueprint = Blueprint(resources=[res.Database(name="DB2")], scope=BlueprintScope.SCHEMA)
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    with pytest.raises(NonConformingPlanException):
        blueprint._raise_for_nonconforming_plan(session_ctx, plan)


def test_blueprint_plan_scope_stubbing(session_ctx):
    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:database/DB1"): {"owner": "SYSADMIN"},
        parse_URN("urn::ABCD123:schema/DB1.PUBLIC"): {"owner": "SYSADMIN"},
    }

    blueprint = Blueprint(
        resources=[res.Task(name="TASK1")],
        scope=BlueprintScope.SCHEMA,
        database="DB1",
        schema="PUBLIC",
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1

    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:database/DB1"): {"owner": "SYSADMIN"},
        parse_URN("urn::ABCD123:schema/DB1.PUBLIC"): {"owner": "SYSADMIN"},
        parse_URN("urn::ABCD123:schema/DB1.ANOTHER_SCHEMA"): {"owner": "SYSADMIN"},
    }

    blueprint = Blueprint(
        resources=[res.Task(name="TASK1")],
        scope=BlueprintScope.SCHEMA,
        database="DB1",
        schema="ANOTHER_SCHEMA",
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1

    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:database/DB1"): {"owner": "SYSADMIN"},
        parse_URN("urn::ABCD123:schema/DB1.PUBLIC"): {"owner": "SYSADMIN"},
    }

    blueprint = Blueprint(
        resources=[res.Schema(name="A_THIRD_SCHEMA"), res.Task(name="TASK1")],
        scope=BlueprintScope.SCHEMA,
        database="DB1",
        schema="A_THIRD_SCHEMA",
    )
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 2
